use std::{fmt::Display, str::FromStr};

use tracing_subscriber::EnvFilter;

#[derive(Debug)]
pub enum LogLevel {
	Trace,
	Debug,
	Info,
	Warn,
	Error,
}

impl Default for LogLevel {
	fn default() -> Self {
		Self::Info
	}
}

impl Display for LogLevel {
	fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
		match self {
			Self::Trace => write!(f, "trace"),
			Self::Debug => write!(f, "debug"),
			Self::Info => write!(f, "info"),
			Self::Warn => write!(f, "warn"),
			Self::Error => write!(f, "error"),
		}
	}
}

#[derive(Debug, Clone, Copy)]
pub enum LoggingPreset {
	Silent,
	Quiet,
	Default,
	Verbose,
	Trace,
}

impl Default for LoggingPreset {
	fn default() -> Self {
		return Self::Default;
	}
}

impl LoggingPreset {
	pub const fn from_number(level: i32) -> Self {
		if level >= 2 {
			Self::Trace
		} else if level == 1 {
			Self::Verbose
		} else if level == 0 {
			Self::Default // 0 should always be default
		} else if level == -1 {
			Self::Quiet
		} else {
			Self::Silent
		}
	}

	pub const fn get_config(&self) -> LoggingConfig {
		match self {
			Self::Silent => LoggingConfig {
				other: LogLevel::Error,

				galactica: LogLevel::Error,
			},

			Self::Quiet => LoggingConfig {
				other: LogLevel::Warn,

				galactica: LogLevel::Warn,
			},

			Self::Default => LoggingConfig {
				other: LogLevel::Warn,

				galactica: LogLevel::Info,
			},

			Self::Verbose => LoggingConfig {
				other: LogLevel::Warn,

				galactica: LogLevel::Debug,
			},

			Self::Trace => LoggingConfig {
				other: LogLevel::Trace,

				galactica: LogLevel::Trace,
			},
		}
	}
}

pub struct LoggingConfig {
	other: LogLevel,

	galactica: LogLevel,
}

impl From<LoggingConfig> for EnvFilter {
	fn from(conf: LoggingConfig) -> Self {
		EnvFilter::from_str(
			&[
				//
				// Configurable sources
				//
				// // Edged
				format!("galactica={}", conf.galactica),
				conf.other.to_string(),
			]
			.join(","),
		)
		.unwrap()
	}
}
