use serde::Deserialize;
use std::{collections::HashMap, sync::Arc};

use crate::{
	error::{ContentLoadError, ContentLoadResult},
	Content, ContentBuildContext, ContentIndex,
};

pub(crate) mod syntax {
	use std::collections::HashMap;

	use serde::Deserialize;

	use crate::ContentIndex;
	// Raw serde syntax structs.
	// These are never seen by code outside this crate.

	#[derive(Debug, Deserialize)]
	pub struct Faction {
		pub display_name: String,
		pub color: [f32; 3],
		pub relationship: HashMap<ContentIndex, super::Relationship>,
	}
}

/// How two factions should interact with each other.
/// Relationships are directional: the relationship of
/// `a` to `b` may not equal the relationship of `b` to `a`.
///
/// Relationships dictate how a ship of THIS faction
/// will interact with a ship of the OTHER faction.
#[derive(Debug, Deserialize, Clone, Copy)]
pub enum Relationship {
	/// Attack this faction
	#[serde(rename = "hostile")]
	Hostile,

	/// Ignore this faction
	#[serde(rename = "neutral")]
	Neutral,

	/// Protect this faction
	#[serde(rename = "friend")]
	Friend,
}

/// Represents a game faction
#[derive(Debug, Clone)]
pub struct Faction {
	/// The name of this faction
	pub index: ContentIndex,

	/// The pretty name of this faction
	pub display_name: String,

	/// This faction's color.
	/// Format is RGB, with each color between 0 and 1.
	pub color: [f32; 3],

	/// Relationships between this faction and other factions
	/// This is guaranteed to contain an entry for ALL factions.
	pub relationships: HashMap<ContentIndex, Relationship>,
}

impl crate::Build for Faction {
	type InputSyntaxType = HashMap<String, syntax::Faction>;

	fn build(
		factions: Self::InputSyntaxType,
		_build_context: &mut ContentBuildContext,
		content: &mut Content,
	) -> ContentLoadResult<()> {
		for (faction_name, faction) in &factions {
			// Compute relationships
			let mut relationships = HashMap::new();
			for (other_name, other_faction) in &factions {
				if let Some(r) = faction.relationship.get(&ContentIndex::new(other_name)) {
					relationships.insert(ContentIndex::new(other_name), *r);
				} else {
					// Default relationship, if not specified

					// Look at reverse direction...
					let other = other_faction
						.relationship
						.get(&ContentIndex::new(&faction_name));
					relationships.insert(
						ContentIndex::new(other_name),
						// ... and pick a relationship based on that.
						match other {
							Some(Relationship::Hostile) => Relationship::Hostile {},
							_ => Relationship::Neutral {},
						},
					);
				}
			}

			if faction.color[0] > 1.0
				|| faction.color[0] < 0.0
				|| faction.color[1] > 1.0
				|| faction.color[1] < 0.0
				|| faction.color[2] > 1.0
				|| faction.color[2] < 0.0
			{
				return Err(ContentLoadError::Generic {
					msg: format!(
						"Invalid color for faction `{}`. Value out of range.",
						faction_name
					),
				});
			}

			content.factions.insert(
				ContentIndex::new(faction_name),
				Arc::new(Self {
					index: ContentIndex::new(faction_name),
					display_name: faction.display_name.to_owned(),
					relationships,
					color: faction.color,
				}),
			);
		}

		return Ok(());
	}
}
