use lazy_static::lazy_static;
use std::{collections::HashMap, sync::Arc};
use tracing::{debug, info};

use crate::{
	error::{ContentLoadError, ContentLoadResult},
	Content, ContentBuildContext, ContentIndex,
};

pub(crate) mod syntax {
	use crate::{
		error::{ContentLoadError, ContentLoadResult},
		Content, ContentIndex,
	};
	use serde::Deserialize;
	use std::{collections::HashMap, path::PathBuf, sync::Arc};

	// Raw serde syntax structs.
	// These are never seen by code outside this crate.

	/// Convenience variants of sprite definitions
	#[derive(Debug, Deserialize)]
	#[serde(untagged)]
	pub enum Sprite {
		Static(StaticSprite),
		OneSection(SpriteSection),
		Complete(CompleteSprite),
	}

	/// Two ways to specify animation length
	#[derive(Debug, Deserialize)]
	pub enum TimingVariant {
		/// The duration of this whole section
		#[serde(rename = "duration")]
		Duration(f32),

		/// The fps of this section
		#[serde(rename = "fps")]
		Fps(f32),
	}

	#[derive(Debug, Deserialize)]
	pub struct Timing {
		#[serde(flatten)]
		pub variant: TimingVariant,
	}

	/// An unanimated sprite
	#[derive(Debug, Deserialize)]
	pub struct StaticSprite {
		pub file: PathBuf,
	}

	/// The proper, full sprite definition
	#[derive(Debug, Deserialize)]
	pub struct CompleteSprite {
		pub section: HashMap<ContentIndex, SpriteSection>,
		pub start_at: SectionEdge,
	}

	/// A single animation section
	#[derive(Debug, Deserialize)]
	pub struct SpriteSection {
		pub frames: Vec<PathBuf>,
		pub timing: Timing,
		pub top: Option<SectionEdge>,
		pub bot: Option<SectionEdge>,
	}

	impl SpriteSection {
		pub fn resolve_edges(
			&self,
			sections: &HashMap<ContentIndex, Arc<super::SpriteSection>>,
			sec: &mut super::SpriteSection,
		) -> ContentLoadResult<()> {
			let edge_top = match &self.top {
				Some(x) => super::resolve_edge_as_edge(sections, &x.val, sec.frame_duration)?,
				None => super::SectionEdge::Stop,
			};

			let edge_bot = match &self.bot {
				Some(x) => super::resolve_edge_as_edge(sections, &x.val, sec.frame_duration)?,
				None => super::SectionEdge::Stop,
			};

			sec.edge_bot = edge_bot;
			sec.edge_top = edge_top;

			return Ok(());
		}

		pub fn add_to(
			&self,
			ct: &mut Content,
		) -> ContentLoadResult<((u32, u32), super::SpriteSection)> {
			// Make sure all frames have the same size and add them
			// to the frame vector
			let mut dim = None;
			let mut frames: Vec<u32> = Vec::new();
			for f in &self.frames {
				let idx = match ct.sprite_atlas.get_idx_by_path(f) {
					Some(s) => s,
					None => {
						return Err(ContentLoadError::Generic {
							msg: format!("file `{}` isn't in the sprite atlas", f.display()),
						});
					}
				};
				let img = &ct.sprite_atlas.get_by_idx(idx);

				match dim {
					None => dim = Some(img.true_size),
					Some(e) => {
						if img.true_size != e {
							return Err(ContentLoadError::Generic {
								msg: "failed to load section frames because frames have different sizes".to_string(),
							});
						}
					}
				}

				frames.push(img.idx.into());
			}

			if frames.len() == 0 {
				return Err(ContentLoadError::Generic {
					msg: "sprite sections must not be empty".to_string(),
				});
			}

			let dim = dim.unwrap();

			let frame_duration = match self.timing.variant {
				TimingVariant::Duration(d) => d / self.frames.len() as f32,
				TimingVariant::Fps(f) => 1.0 / f,
			};

			if frame_duration <= 0.0 {
				return Err(ContentLoadError::Generic {
					msg: "frame duration must be positive (and therefore nonzero)".to_string(),
				});
			}

			return Ok((
				dim,
				super::SpriteSection {
					frames,
					frame_duration,
					// These are changed later, after all sections are built
					edge_top: super::SectionEdge::Stop,
					edge_bot: super::SectionEdge::Stop,
				},
			));
		}
	}

	/// A link between two animation sections
	#[derive(Debug, Deserialize)]
	#[serde(transparent)]
	pub struct SectionEdge {
		pub val: String,
	}
}

/// An edge between two animation sections
#[derive(Debug, Clone)]
pub enum SectionEdge {
	/// Stop at the last frame of this section
	Stop,

	/// Play the given section from the bottm
	Bot {
		/// The section to play
		section: Arc<SpriteSection>,

		/// The length of this edge, in seconds
		duration: f32,
	},

	/// Play the given section from the top
	Top {
		/// The section to play
		section: Arc<SpriteSection>,

		/// The length of this edge, in seconds
		duration: f32,
	},

	/// Replay this section in the opposite direction
	Reverse {
		/// The length of this edge, in seconds
		duration: f32,
	},

	/// Restart this section from the opposite end
	Repeat {
		/// The length of this edge, in seconds
		duration: f32,
	},
}

/// Where to start an animation
#[derive(Debug, Clone)]
pub enum StartEdge {
	/// Play the given section from the bottm
	Bot {
		/// The section to play
		section: Arc<SpriteSection>,
	},

	/// Play the given section from the top
	Top {
		/// The section to play
		section: Arc<SpriteSection>,
	},
}

impl Into<SectionEdge> for StartEdge {
	fn into(self) -> SectionEdge {
		match self {
			Self::Bot { section } => SectionEdge::Bot {
				section,
				duration: 0.0,
			},
			Self::Top { section } => SectionEdge::Top {
				section,
				duration: 0.0,
			},
		}
	}
}

/// Represents a sprite that may be used in the game.
#[derive(Debug, Clone)]
pub struct Sprite {
	/// This object's index
	pub index: ContentIndex,

	/// Where this sprite starts playing
	pub start_at: StartEdge,

	/// This sprite's animation sections
	pub sections: HashMap<ContentIndex, Arc<SpriteSection>>,

	/// Aspect ratio of this sprite (width / height)
	pub aspect: f32,
}

lazy_static! {
	/// The universal "hidden" sprite section, available in all sprites.
	/// A SpriteAutomaton in this section will not be drawn.
	pub static ref HIDDEN_SPRITE_SECTION: Arc<SpriteSection> = Arc::new(SpriteSection {
		frames: vec![0],
		frame_duration: 0.0,
		edge_bot: SectionEdge::Stop,
		edge_top: SectionEdge::Stop,
	});
}

impl Sprite {
	/// Get the index of the texture of this sprite's first frame
	pub fn get_first_frame(&self) -> u32 {
		match &self.start_at {
			StartEdge::Bot { section } => *section.frames.last().unwrap(),
			StartEdge::Top { section } => *section.frames.first().unwrap(),
		}
	}

	/// Get this sprite's starting section
	pub fn get_start_section(&self) -> Arc<SpriteSection> {
		match &self.start_at {
			StartEdge::Bot { section } => section.clone(),
			StartEdge::Top { section } => section.clone(),
		}
	}

	/// Iterate this sprite's sections
	pub fn iter_sections(&self) -> impl Iterator<Item = &Arc<SpriteSection>> {
		self.sections.values()
	}
}

/// A part of a sprite's animation
#[derive(Debug, Clone)]
pub struct SpriteSection {
	/// The texture index of each frame in this animation section.
	/// unanimated sections have one frame.
	pub frames: Vec<u32>,

	/// The speed of this sprite's animation.
	/// This must always be positive (and therefore, nonzero)
	pub frame_duration: f32,

	/// What to do when we reach the top of this section
	pub edge_top: SectionEdge,

	/// What to do when we reach the bottom of this section
	pub edge_bot: SectionEdge,
}

/// Resolve an edge specification string as a StartEdge
pub fn resolve_edge_as_start(
	sections: &HashMap<ContentIndex, Arc<SpriteSection>>,
	edge_string: &str,
) -> ContentLoadResult<super::StartEdge> {
	debug!(message = "Resolving start edge");
	let e = resolve_edge_as_edge(sections, edge_string, 0.0)?;
	match e {
		super::SectionEdge::Bot { section, .. } => Ok(super::StartEdge::Bot { section }),
		super::SectionEdge::Top { section, .. } => Ok(super::StartEdge::Top { section }),
		_ => {
			return Err(ContentLoadError::Generic {
				msg: format!("bad section start specification `{}`", edge_string),
			});
		}
	}
}

/// Resolve an edge specifiation string as a SectionEdge
pub fn resolve_edge_as_edge(
	sections: &HashMap<ContentIndex, Arc<SpriteSection>>,
	edge_string: &str,
	duration: f32,
) -> ContentLoadResult<super::SectionEdge> {
	if edge_string == "hidden" {
		return Ok(super::SectionEdge::Top {
			section: crate::HIDDEN_SPRITE_SECTION.clone(),
			duration,
		});
	}

	if edge_string == "stop" {
		return Ok(super::SectionEdge::Stop);
	}

	if edge_string == "reverse" {
		return Ok(super::SectionEdge::Reverse { duration });
	}

	if edge_string == "repeat" {
		return Ok(super::SectionEdge::Repeat { duration });
	}

	let (section_name, start_point) = match edge_string.split_once(":") {
		Some(x) => x,
		None => {
			return Err(ContentLoadError::Generic {
				msg: format!("bad section edge specification `{}`", edge_string),
			});
		}
	};

	let section = match sections.get(&ContentIndex::new(section_name)) {
		Some(s) => s,
		None => {
			return Err(ContentLoadError::Generic {
				msg: format!("section `{}` doesn't exist", section_name),
			});
		}
	};

	match start_point {
		"top" => Ok(super::SectionEdge::Top {
			section: section.clone(),
			duration,
		}),
		"bot" => Ok(super::SectionEdge::Bot {
			section: section.clone(),
			duration,
		}),
		_ => {
			return Err(ContentLoadError::Generic {
				msg: format!("invalid target `{}`", start_point),
			});
		}
	}
}

impl crate::Build for Sprite {
	type InputSyntaxType = HashMap<String, syntax::Sprite>;

	fn build(
		sprites: Self::InputSyntaxType,
		_build_context: &mut ContentBuildContext,
		ct: &mut Content,
	) -> ContentLoadResult<()> {
		for (sprite_name, t) in sprites {
			info!(message = "Parsing sprite", sprite = sprite_name);

			match t {
				syntax::Sprite::Static(t) => {
					let idx = match ct.sprite_atlas.get_idx_by_path(&t.file) {
						Some(s) => s,
						None => {
							return Err(ContentLoadError::Generic {
								msg: format!(
									"file `{}` isn't in the sprite atlas, cannot proceed",
									t.file.display()
								),
							});
						}
					};
					let img = &ct.sprite_atlas.get_by_idx(idx);
					let aspect = img.w / img.h;

					let section = Arc::new(SpriteSection {
						frames: vec![img.idx.into()],
						// We implement unanimated sprites with a very fast framerate
						// and STOP endpoints.
						frame_duration: 0.01,
						edge_top: SectionEdge::Stop,
						edge_bot: SectionEdge::Stop,
					});

					let sprite = Arc::new(Self {
						index: ContentIndex::new(&sprite_name),
						start_at: StartEdge::Top {
							section: section.clone(),
						},
						sections: {
							let mut h = HashMap::new();
							h.insert(ContentIndex::new("anim"), section);
							h
						},
						aspect,
					});
					ct.sprites.insert(sprite.index.clone(), sprite);
				}
				syntax::Sprite::OneSection(s) => {
					let (dim, section) = s.add_to(ct)?;
					let aspect = dim.0 as f32 / dim.1 as f32;

					let section = Arc::new(section);

					let sprite = Arc::new(Self {
						index: ContentIndex::new(&sprite_name),
						start_at: StartEdge::Top {
							section: section.clone(),
						},
						sections: {
							let mut h = HashMap::new();
							h.insert(ContentIndex::new("anim"), section.clone());
							h
						},
						aspect,
					});
					ct.sprites.insert(sprite.index.clone(), sprite);
				}
				syntax::Sprite::Complete(s) => {
					let mut sections = HashMap::new();

					let mut dim = None;

					for (name, sec) in &s.section {
						info!(
							message = "Parsing sprite section",
							sprite = sprite_name,
							section = ?name
						);

						let (d, s) = sec.add_to(ct)?;

						// Make sure all dimensions are the same
						if dim.is_none() {
							dim = Some(d);
						} else if dim.unwrap() != d {
							return Err(ContentLoadError::Generic {
								msg: format!(
									"could not load sprite, image sizes in section `{}` are different",
								name
								),
							});
						}

						sections.insert(name.clone(), Arc::new(s));
					}

					// We need to clone this here, thanks to self-reference.
					let tmp_sections = sections.clone();

					for (name, sec) in &s.section {
						let parsed_sec = sections.get_mut(name).unwrap();
						let parsed_sec = Arc::make_mut(parsed_sec);
						sec.resolve_edges(&tmp_sections, parsed_sec)?;
					}

					let dim = dim.unwrap();
					let aspect = dim.0 as f32 / dim.1 as f32;

					let start_at = resolve_edge_as_start(&sections, &s.start_at.val)?;

					let sprite = Arc::new(Self {
						index: ContentIndex::new(&sprite_name),
						start_at,
						sections,
						aspect,
					});
					ct.sprites.insert(sprite.index.clone(), sprite);
				}
			}
		}

		return Ok(());
	}
}
