use galactica_util::to_radians;
use nalgebra::{Point2, Point3};
use std::{
	collections::{HashMap, HashSet},
	sync::Arc,
};
use tracing::info;

use crate::{
	error::{ContentLoadError, ContentLoadResult},
	util::Polar,
	Content, ContentBuildContext, ContentIndex, Outfit, Sprite,
};

pub(crate) mod syntax {
	use serde::Deserialize;
	use std::collections::HashMap;

	use crate::ContentIndex;
	// Raw serde syntax structs.
	// These are never seen by code outside this crate.

	#[derive(Debug, Deserialize)]
	pub struct System {
		pub name: String,
		pub object: HashMap<ContentIndex, Object>,
	}

	#[derive(Debug, Deserialize)]
	pub struct Object {
		pub sprite: ContentIndex,
		pub position: Position,
		pub size: f32,
		pub radius: Option<f32>,
		pub angle: Option<f32>,
		pub landable: Option<Landable>,
	}

	#[derive(Debug, Deserialize)]
	pub struct Landable {
		pub name: String,
		pub desc: String,
		pub image: ContentIndex,
		pub outfitter: Vec<ContentIndex>,
	}

	#[derive(Debug, Deserialize)]
	#[serde(untagged)]
	pub enum Position {
		Polar(PolarCoords),
		Cartesian(CoordinatesThree),
	}

	#[derive(Debug, Deserialize)]
	pub struct PolarCoords {
		pub center: CoordinatesTwo,
		pub radius: f32,
		pub angle: f32,
		pub z: f32,
	}

	#[derive(Debug, Deserialize)]
	#[serde(untagged)]
	pub enum CoordinatesTwo {
		Label(ContentIndex),
		Coords([f32; 2]),
	}

	impl ToString for CoordinatesTwo {
		fn to_string(&self) -> String {
			match self {
				Self::Label(s) => s.to_string(),
				Self::Coords(v) => format!("{:?}", v),
			}
		}
	}

	#[derive(Debug, Deserialize)]
	#[serde(untagged)]
	pub enum CoordinatesThree {
		Label(ContentIndex),
		Coords([f32; 3]),
	}

	impl ToString for CoordinatesThree {
		fn to_string(&self) -> String {
			match self {
				Self::Label(s) => s.to_string(),
				Self::Coords(v) => format!("{:?}", v),
			}
		}
	}
}

// Processed data structs.
// These are exported.

/// Represents a star system
#[derive(Debug, Clone)]
pub struct System {
	/// This object's name
	pub display_name: String,

	/// This object's index
	pub index: ContentIndex,

	/// Objects in this system
	pub objects: HashMap<ContentIndex, Arc<SystemObject>>,
}

/// Represents an orbiting body in a star system
/// (A star, planet, moon, satellite, etc)
/// These may be landable and may be decorative.
/// System objects to not interact with the physics engine.
#[derive(Debug, Clone)]
pub struct SystemObject {
	/// This object's index
	pub index: ContentIndex,

	/// This object's sprite
	pub sprite: Arc<Sprite>,

	/// This object's size.
	/// Measured as height in game units.
	/// This value is scaled for distance
	/// (i.e, the z-value of position)
	pub size: f32,

	/// This object's position, in game coordinates,
	/// relative to the system's center (0, 0).
	pub pos: Point3<f32>,

	/// This object's sprite's angle, in radians
	pub angle: f32,

	/// If true, ships may land on this object
	pub landable: Option<LandableSystemObject>,
}

#[derive(Debug, Clone)]
pub struct LandableSystemObject {
	/// This object's name
	pub display_name: String,

	/// The description of this object
	pub desc: String,

	/// This object's image
	pub image: Arc<Sprite>,

	/// The outfits we can buy here
	/// If this is empty, this landable has no outfitter.
	pub outfitter: Vec<Arc<Outfit>>,
}

/// Helper function for resolve_position, never called on its own.
fn resolve_coordinates(
	objects: &HashMap<ContentIndex, syntax::Object>,
	cor: &syntax::CoordinatesThree,
	mut cycle_detector: HashSet<ContentIndex>,
) -> ContentLoadResult<Point3<f32>> {
	match cor {
		syntax::CoordinatesThree::Coords(c) => Ok((*c).into()),
		syntax::CoordinatesThree::Label(l) => {
			if cycle_detector.contains(l) {
				return Err(ContentLoadError::Generic {
					msg: format!(
						"Found coordinate cycle: `{}`",
						cycle_detector.iter().fold(String::new(), |sum, a| {
							if sum.is_empty() {
								a.to_string()
							} else {
								format!("{sum} -> {a}")
							}
						})
					),
				});
			}
			cycle_detector.insert(l.clone());

			let p = match objects.get(l) {
				Some(p) => p,
				None => {
					return Err(ContentLoadError::Generic {
						msg: format!("Could not resolve coordinate label `{l}`"),
					});
				}
			};
			Ok(resolve_position(&objects, &p, cycle_detector)?)
		}
	}
}

/// Given an object, resolve its position as a Point3.
fn resolve_position(
	objects: &HashMap<ContentIndex, syntax::Object>,
	obj: &syntax::Object,
	cycle_detector: HashSet<ContentIndex>,
) -> ContentLoadResult<Point3<f32>> {
	match &obj.position {
		syntax::Position::Cartesian(c) => Ok(resolve_coordinates(objects, &c, cycle_detector)?),
		syntax::Position::Polar(p) => {
			let three = match &p.center {
				syntax::CoordinatesTwo::Label(s) => syntax::CoordinatesThree::Label(s.clone()),
				syntax::CoordinatesTwo::Coords(v) => {
					syntax::CoordinatesThree::Coords([v[0], v[1], f32::NAN])
				}
			};
			let r = resolve_coordinates(&objects, &three, cycle_detector)?;
			let plane = Polar {
				center: Point2::new(r.x, r.y),
				radius: p.radius,
				angle: to_radians(p.angle),
			}
			.to_cartesian();
			Ok(Point3::new(plane.x, plane.y, p.z))
		}
	}
}

impl crate::Build for System {
	type InputSyntaxType = HashMap<String, syntax::System>;

	fn build(
		system: Self::InputSyntaxType,
		_build_context: &mut ContentBuildContext,
		content: &mut Content,
	) -> ContentLoadResult<()> {
		for (system_name, system) in system {
			info!(message = "Parsing system", system = system_name);

			let mut objects = HashMap::new();

			for (index, object) in &system.object {
				info!(
					message = "Parsing system object",
					system = system_name,
					object_idx = ?index
				);

				let mut cycle_detector = HashSet::new();
				cycle_detector.insert(index.clone());

				let sprite = match content.sprites.get(&object.sprite) {
					None => {
						return Err(ContentLoadError::Generic {
							msg: format!("sprite `{}` doesn't exist", object.sprite),
						});
					}

					Some(t) => t.clone(),
				};

				let landable = 'landable: {
					if object.landable.is_none() {
						break 'landable None;
					}
					let l = object.landable.as_ref().unwrap();

					let image = match content.sprites.get(&l.image) {
						None => {
							return Err(ContentLoadError::Generic {
								msg: format!("sprite `{}` doesn't exist", object.sprite),
							});
						}

						Some(t) => t.clone(),
					};

					let mut outfitter = Vec::new();
					for o in &l.outfitter {
						match content.outfits.get(&o) {
							Some(x) => outfitter.push(x.clone()),
							None => {
								return Err(ContentLoadError::Generic {
									msg: format!("outfit `{o}` doesn't exist"),
								});
							}
						}
					}

					break 'landable Some(LandableSystemObject {
						image,
						outfitter,
						display_name: l.name.clone(),
						// TODO: better linebreaks, handle double spaces
						// Tabs
						desc: l.desc.replace("\n", " ").replace("<br>", "\n"),
					});
				};

				objects.insert(
					index.clone(),
					Arc::new(SystemObject {
						index: index.clone(),
						sprite,
						pos: resolve_position(&system.object, &object, cycle_detector)?,
						size: object.size,
						angle: to_radians(object.angle.unwrap_or(0.0)),
						landable,
					}),
				);
			}

			content.systems.insert(
				ContentIndex::new(&system_name),
				Arc::new(Self {
					index: ContentIndex::new(&system_name),
					display_name: system.name,
					objects,
				}),
			);
		}

		return Ok(());
	}
}
