//! This crate creates texture atlases from an asset tree.

mod atlasset;
pub use atlasset::*;

use std::{
	collections::HashMap,
	num::NonZeroU32,
	path::{Path, PathBuf},
};

use serde::{Deserialize, Serialize};

/// The location of a single image in a sprite atlas
#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct SpriteAtlasImage {
	/// The index of the atlas this image is in
	/// This is an index in SpriteAtlas.atlas_list
	pub atlas: u32,

	/// A globally unique, consecutively numbered index for this sprite.
	/// This is nonzero because index zero is reserved for the "hidden" texture.
	pub idx: NonZeroU32,

	/// The size of this image, in pixels
	pub true_size: (u32, u32),

	/// x-position of this image
	/// (between 0 and 1, using wgpu texture coordinates)
	pub x: f32,

	/// y-position of this image
	/// (between 0 and 1, using wgpu texture coordinates)
	pub y: f32,

	/// Width of this image
	/// (between 0 and 1, using wgpu texture coordinates)
	pub w: f32,

	/// Height of this image
	/// (between 0 and 1, using wgpu texture coordinates)
	pub h: f32,
}

/// A map between file paths (relative to the root asset dir)
/// and [`AtlasTexture`]s.
#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct SpriteAtlas {
	/// The images in this atlas
	pub(crate) index: Vec<SpriteAtlasImage>,

	/// Map paths to image indices
	path_map: HashMap<PathBuf, NonZeroU32>,

	/// The file names of the atlas textures we've generated
	pub atlas_list: Vec<String>,
}

impl SpriteAtlas {
	/// Make an empty [`SpriteAtlasIndex`]
	pub fn new() -> Self {
		Self {
			path_map: HashMap::new(),
			index: Vec::new(),
			atlas_list: Vec::new(),
		}
	}

	/// Get a SpriteAtlasImage by index
	pub fn get_by_idx(&self, idx: NonZeroU32) -> &SpriteAtlasImage {
		&self.index[idx.get() as usize - 1]
	}

	/// Get an image index from its path
	/// returns None if this path isn't in this index
	pub fn get_idx_by_path(&self, path: &Path) -> Option<NonZeroU32> {
		self.path_map.get(path).map(|x| *x)
	}

	/// Iterate all images in this atlas
	pub fn iter_images(&self) -> impl Iterator<Item = &SpriteAtlasImage> {
		self.index.iter()
	}

	/// Get the number of images in this atlas
	pub fn len(&self) -> u32 {
		self.index.len() as u32
	}

	/// Add an image with the given path to this index
	pub fn push(&mut self, p: &Path, i: SpriteAtlasImage) {
		self.path_map.insert(
			p.to_path_buf(),
			NonZeroU32::new(self.index.len() as u32 + 1).unwrap(),
		);

		self.index.push(i);
	}
}
