use galactica_content::{Effect, EffectVelocity, SpriteAutomaton};
use nalgebra::{Point2, Rotation2, Vector2};
use rand::Rng;
use rapier2d::dynamics::{RevoluteJointBuilder, RigidBodyBuilder, RigidBodyHandle, RigidBodyType};
use std::sync::Arc;

use crate::physwrapper::PhysWrapper;

/// An instance of an effect in a simulation
#[derive(Debug, Clone)]
pub struct EffectInstance {
	/// The sprite to use for this effect
	pub anim: SpriteAutomaton,

	/// This effect's velocity, in world coordinates
	pub rigid_body: RigidBodyHandle,

	/// This effect's lifetime, in seconds
	lifetime: f32,

	/// The size of this effect,
	/// given as height in world units.
	pub size: f32,

	/// Fade this effect over this many seconds as it expires
	pub fade: f32,

	/// If true, this effect has been destroyed,
	/// and needs to be removed from the game.
	is_destroyed: bool,
}

impl EffectInstance {
	/// Create a new effect inside `Wrapper`
	pub fn new(
		wrapper: &mut PhysWrapper,
		effect: Arc<Effect>,
		// Where to spawn the particle, in world space.
		pos: Vector2<f32>,
		parent: RigidBodyHandle,
		target: Option<RigidBodyHandle>,
	) -> Self {
		let mut rng = rand::thread_rng();
		let parent_body = wrapper.get_rigid_body(parent).unwrap();
		let parent_angle = parent_body.rotation().angle();
		let parent_pos = *parent_body.translation();
		let parent_velocity = parent_body.velocity_at_point(parent_body.center_of_mass());

		let angvel = if effect.angvel_rng == 0.0 {
			effect.angvel
		} else {
			effect.angvel + rng.gen_range(-effect.angvel_rng..=effect.angvel_rng)
		};
		let angle = if effect.angle_rng == 0.0 {
			parent_angle + effect.angle
		} else {
			parent_angle + effect.angle + rng.gen_range(-effect.angle_rng..=effect.angle_rng)
		};

		let target_velocity = {
			if let Some(target) = target {
				let target_body = wrapper.get_rigid_body(target).unwrap();
				target_body.velocity_at_point(&Point2::new(pos.x, pos.y))
			} else {
				Vector2::new(0.0, 0.0)
			}
		};

		match effect.velocity {
			EffectVelocity::StickyTarget | EffectVelocity::StickyParent => {
				let rigid_body = wrapper.insert_rigid_body(
					RigidBodyBuilder::new(RigidBodyType::Dynamic)
						.additional_mass(f32::MIN_POSITIVE)
						.position(pos.into())
						.rotation(angle)
						.angvel(angvel)
						.build(),
				);

				match effect.velocity {
					EffectVelocity::StickyParent => {
						let d = Rotation2::new(-parent_angle) * (pos - parent_pos);

						wrapper.add_joint(
							rigid_body,
							parent,
							RevoluteJointBuilder::new()
								.local_anchor1(Point2::new(0.0, 0.0))
								.local_anchor2(Point2::new(d.x, d.y)),
						)
					}
					EffectVelocity::StickyTarget => {
						if target.is_some() {
							let target_body = wrapper.get_rigid_body(target.unwrap()).unwrap();

							// Correct for rotation, since joint coordinates are relative
							// and input coordinates are in world space.
							let d = Rotation2::new(-target_body.rotation().angle())
								* (pos - target_body.translation());

							wrapper.add_joint(
								rigid_body,
								target.unwrap(),
								RevoluteJointBuilder::new()
									.local_anchor1(Point2::new(0.0, 0.0))
									.local_anchor2(Point2::new(d.x, d.y)),
							)
						}
					}
					_ => unreachable!("lol what?"),
				};

				EffectInstance {
					anim: SpriteAutomaton::new(effect.sprite.clone()),
					rigid_body,
					lifetime: 0f32.max(
						effect.lifetime + rng.gen_range(-effect.lifetime_rng..=effect.lifetime_rng),
					),

					// Make sure size isn't negative. This check should be on EVERY rng!
					size: 0f32.max(effect.size + rng.gen_range(-effect.size_rng..=effect.size_rng)),
					fade: 0f32.max(effect.fade + rng.gen_range(-effect.fade_rng..=effect.fade_rng)),
					is_destroyed: false,
				}
			}

			EffectVelocity::Explicit {
				scale_parent,
				scale_parent_rng,
				scale_target,
				scale_target_rng,
				direction_rng,
			} => {
				let velocity = {
					let a = rng.gen_range(-scale_parent_rng..=scale_parent_rng);
					let b = rng.gen_range(-scale_target_rng..=scale_target_rng);

					let velocity = ((scale_parent + a) * parent_velocity)
						+ ((scale_target + b) * target_velocity);

					Rotation2::new(rng.gen_range(-direction_rng..=direction_rng)) * velocity
				};

				let rigid_body = wrapper.insert_rigid_body(
					RigidBodyBuilder::new(RigidBodyType::KinematicVelocityBased)
						.position(pos.into())
						.rotation(angle)
						.angvel(angvel)
						.linvel(velocity)
						.build(),
				);

				EffectInstance {
					anim: SpriteAutomaton::new(effect.sprite.clone()),
					rigid_body,
					lifetime: 0f32.max(
						effect.lifetime + rng.gen_range(-effect.lifetime_rng..=effect.lifetime_rng),
					),

					// Make sure size isn't negative. This check should be on EVERY rng!
					size: 0f32.max(effect.size + rng.gen_range(-effect.size_rng..=effect.size_rng)),
					fade: 0f32.max(effect.fade + rng.gen_range(-effect.fade_rng..=effect.fade_rng)),
					is_destroyed: false,
				}
			}
		}
	}

	/// Step this effect's state by `dt` seconds
	pub fn step(&mut self, dt: f32, wrapper: &mut PhysWrapper) {
		if self.is_destroyed {
			return;
		}

		self.anim.step(dt);
		self.lifetime -= dt;

		if self.lifetime <= 0.0 {
			wrapper.remove_rigid_body(self.rigid_body);
			self.is_destroyed = true;
		}
	}

	/// Should this effect be deleted?
	pub fn is_destroyed(&self) -> bool {
		self.is_destroyed
	}

	/// The remaining lifetime of this effect, in seconds
	pub fn remaining_lifetime(&self) -> f32 {
		self.lifetime
	}
}
