use std::{collections::HashMap, sync::Arc};

use galactica_content::{ContentIndex, GunPoint, Outfit, OutfitSpace, OutfitStats};

/// Possible outcomes when adding an outfit
pub enum OutfitAddResult {
	/// An outfit was successfully added
	Ok,

	/// An outfit could not be added because we don't have enough free space.
	/// The string tells us what kind of space we need:
	/// `outfit,` `weapon,` `engine,` etc. Note that these sometimes overlap:
	/// outfits may need outfit AND weapon space. In these cases, this result
	/// should name the "most specific" kind of space we lack.
	NotEnoughSpace(String),

	/// An outfit couldn't be added because there weren't enough points for it
	/// (e.g, gun points, turret points, etc)
	NotEnoughPoints(String),
}

/// Possible outcomes when removing an outfit
pub enum OutfitRemoveResult {
	/// This outfit was successfully removed
	Ok,

	/// This outfit isn't in this set
	NotExist,
	// TODO:
	// This is where we'll add non-removable outfits,
	// outfits that provide space, etc
}

/// A simple data class, used to keep track of delayed shield generators
#[derive(Debug, Clone)]
pub(crate) struct ShieldGenerator {
	pub outfit: Arc<Outfit>,
	pub delay: f32,
	pub generation: f32,
}

/// This struct keeps track of a ship's outfit loadout.
/// This is a fairly static data structure: it does not keep track of cooldowns,
/// shield damage, etc. It only provides an interface for static stats which are
/// then used elsewhere.
#[derive(Debug, Clone)]
pub struct OutfitSet {
	/// What outfits does this statsum contain?
	outfits: HashMap<ContentIndex, (Arc<Outfit>, u32)>,

	/// Space available in this outfitset.
	/// set at creation and never changes.
	total_space: OutfitSpace,

	/// Space used by the outfits in this set.
	/// This may be negative if certain outfits provide space!
	used_space: OutfitSpace,

	/// The gun points available in this ship.
	/// If value is None, this point is free.
	/// if value is Some, this point is taken.
	gun_points: HashMap<GunPoint, Option<Arc<Outfit>>>,

	/// The combined stats of all outfits in this set.
	/// There are two things to note here:
	/// First, shield_delay is always zero. That is handled
	/// seperately, since it is different for every outfit.
	/// Second, shield_generation represents the MAXIMUM POSSIBLE
	/// shield generation, after all delays have expired.
	///
	/// Note that this field isn't strictly necessary... we could compute stats
	/// by iterating over the outfits in this set. We don't want to do this every
	/// frame, though, so we keep track of the total sum here.
	stats: OutfitStats,

	/// All shield generators in this outfit set
	// These can't be summed into one value, since each has a
	// distinct delay.
	shield_generators: Vec<ShieldGenerator>,
}

impl OutfitSet {
	pub(super) fn new(available_space: OutfitSpace, gun_points: &[GunPoint]) -> Self {
		Self {
			outfits: HashMap::new(),
			total_space: available_space,
			used_space: OutfitSpace::new(),
			gun_points: gun_points.iter().map(|x| (x.clone(), None)).collect(),
			stats: OutfitStats::zero(),
			shield_generators: Vec::new(),
		}
	}

	pub(super) fn add(&mut self, o: &Arc<Outfit>) -> OutfitAddResult {
		if !(self.total_space - self.used_space).can_contain(&o.space) {
			return OutfitAddResult::NotEnoughSpace("TODO".to_string());
		}

		// Check and return as fast as possible,
		// BEFORE we make any changes.
		if o.gun.is_some() {
			let mut added = false;
			for (_, outfit) in &mut self.gun_points {
				if outfit.is_none() {
					*outfit = Some(o.clone());
					added = true;
					break;
				}
			}
			if !added {
				return OutfitAddResult::NotEnoughPoints("gun".to_string());
			}
		}

		self.used_space += o.space;
		self.stats.add(&o.stats);

		if o.stats.shield_generation != 0.0 {
			self.shield_generators.push(ShieldGenerator {
				outfit: o.clone(),
				delay: o.stats.shield_delay,
				generation: o.stats.shield_generation,
			});
		}

		if self.outfits.contains_key(&o.index) {
			self.outfits.get_mut(&o.index).unwrap().1 += 1;
		} else {
			self.outfits.insert(o.index.clone(), (o.clone(), 1));
		}

		return OutfitAddResult::Ok;
	}

	pub(super) fn remove(&mut self, o: &Arc<Outfit>) -> OutfitRemoveResult {
		if !self.outfits.contains_key(&o.index) {
			return OutfitRemoveResult::NotExist;
		}

		let n = self.outfits.get_mut(&o.index).unwrap();
		if n.1 == 1u32 {
			self.outfits.remove(&o.index);
		} else {
			self.outfits.get_mut(&o.index).unwrap().1 -= 1;
		}

		if o.gun.is_some() {
			let (_, x) = self
				.gun_points
				.iter_mut()
				.find(|(_, x)| x.is_some() && x.as_ref().unwrap().index == o.index)
				.unwrap();
			*x = None;
		}

		self.used_space -= o.space;
		self.stats.subtract(&o.stats);

		let index = self
			.shield_generators
			.iter()
			.position(|g| g.outfit.index == o.index);
		if let Some(index) = index {
			self.shield_generators.remove(index);
		}

		return OutfitRemoveResult::Ok;
	}
}

// Simple getters to make sure nobody meddles with our internal state
impl OutfitSet {
	/// The number of outfits in this set
	pub fn len(&self) -> u32 {
		self.outfits.iter().map(|(_, (_, x))| x).sum()
	}

	/// Iterate over all outfits
	pub fn iter_outfits(&self) -> impl Iterator<Item = &(Arc<Outfit>, u32)> {
		self.outfits.values()
	}

	/// Iterate over all gun points
	pub fn iter_gun_points(&self) -> impl Iterator<Item = (&GunPoint, &Option<Arc<Outfit>>)> {
		self.gun_points.iter()
	}

	/// Iterate over all shield generators
	pub(crate) fn iter_shield_generators(&self) -> impl Iterator<Item = &ShieldGenerator> {
		self.shield_generators.iter()
	}

	/// Get the outfit attached to the given gun point
	/// Will panic if this gunpoint is not in this outfit set.
	pub fn get_gun(&self, point: &GunPoint) -> Option<Arc<Outfit>> {
		self.gun_points.get(point).unwrap().clone()
	}

	/// Total available outfit space
	pub fn get_total_space(&self) -> &OutfitSpace {
		&self.total_space
	}

	/// Used outfit space
	pub fn get_used_space(&self) -> &OutfitSpace {
		&self.used_space
	}

	/// Number of available (used & free) gun points
	pub fn total_gun_points(&self) -> usize {
		self.gun_points.len()
	}

	/// Number of free gun points
	pub fn free_gun_points(&self) -> usize {
		self.iter_gun_points().filter(|(_, o)| o.is_none()).count()
	}

	/// Does this set contain `count` of `outfit`?
	pub fn has_outfit(&self, outfit: &Arc<Outfit>, mut count: u32) -> bool {
		for i in self.iter_outfits() {
			if count <= 0 {
				return true;
			}
			if i.0.index == outfit.index {
				count -= 1;
			}
		}
		return count <= 0;
	}

	/// Get the combined stats of all outfits in this set.
	/// There are two things to note here:
	/// First, shield_delay is always zero. That is handled
	/// seperately, since it is different for every outfit.
	/// Second, shield_generation represents the MAXIMUM POSSIBLE
	/// shield generation, after all delays have expired.
	pub fn get_stats(&self) -> &OutfitStats {
		&self.stats
	}
}
