use galactica_content::{
	AnimationState, EnginePoint, Faction, GunPoint, Outfit, ProjectileCollider, Ship,
	SpriteAutomaton,
};
use nalgebra::{vector, Point2, Rotation2, Vector2};
use rand::Rng;
use rapier2d::{
	dynamics::{RigidBodyBuilder, RigidBodyHandle},
	geometry::{ColliderBuilder, ColliderHandle, Group, InteractionGroups},
	pipeline::ActiveEvents,
};
use std::sync::Arc;

use super::{
	collapse::ShipCollapseSequence,
	data::{OutfitRemoveResult, ShipData, ShipState},
};
use crate::{
	get_phys_id,
	instance::{EffectInstance, ProjectileInstance},
	physwrapper::PhysWrapper,
	shipagent::{ShipAgent, ShipControls},
	stateframe::{PhysSimShipHandle, ShipStateframe, SystemStateframe},
	NewObjects,
};

/// A ship instance in the system
#[derive(Debug)]
pub struct ShipInstance {
	/// This ship's unique id
	pub uid: u64,

	/// This ship's physics handle
	pub rigid_body: RigidBodyHandle,

	/// This ship's collider
	pub collider: ColliderHandle,

	/// This ship's game data
	pub(crate) data: ShipData,

	/// This ship's sprite animation state
	anim: SpriteAutomaton,

	/// Animation state for each of this ship's engines
	engine_anim: Vec<(EnginePoint, SpriteAutomaton)>,

	/// This ship's controls
	pub(crate) controls: ShipControls,

	/// This ship's controls during the last frame
	last_controls: ShipControls,

	/// This ship's collapse sequence
	collapse_sequence: Option<ShipCollapseSequence>,

	/// If set, the given agent temporarily controls this ship
	autopilot: Option<Box<dyn ShipAgent>>,

	/// The agent that controls this ship when no autopilot is enabled
	agent: Box<dyn ShipAgent>,

	/// If true, this ship's collider has been destroyed,
	/// and this struct is waiting to be cleaned up.
	///
	/// Note that this is NOT "in-game" destroyed,
	/// but rather "internal game logic" destroyed.
	/// In-game destroyed corresponds to the "Dead" state.
	is_destroyed: bool,
}

impl ShipInstance {
	/// Make a new ship
	pub(crate) fn new(
		ship: Arc<Ship>,
		faction: Arc<Faction>,
		agent: Box<dyn ShipAgent>,
		rigid_body: RigidBodyHandle,
		collider: ColliderHandle,
	) -> Self {
		ShipInstance {
			uid: get_phys_id(),
			anim: SpriteAutomaton::new(ship.sprite.clone()),
			rigid_body,
			collider,
			data: ShipData::new(ship.clone(), faction),
			engine_anim: Vec::new(),
			controls: ShipControls::new(),
			last_controls: ShipControls::new(),
			collapse_sequence: Some(ShipCollapseSequence::new(ship.collapse.length)),
			is_destroyed: false,
			autopilot: None,
			agent,
		}
	}

	/// Return a snapshot of this ship's state
	pub fn get_stateframe(&self, phys: &PhysWrapper) -> ShipStateframe {
		return ShipStateframe {
			uid: self.uid,
			data: self.data.clone(),
			anim: self.anim.clone(),
			engine_anim: self.engine_anim.clone(),
			collapse_sequence: self.collapse_sequence.clone(),
			rigidbody: phys.get_rigid_body(self.rigid_body).unwrap().clone(),
		};
	}

	/// Step this ship's state by t seconds
	/// pub in phys
	pub(crate) fn step(
		&mut self,
		dt: f32,
		sf: &SystemStateframe,
		wrapper: &mut PhysWrapper,
		new: &mut NewObjects,
	) {
		if self.is_destroyed {
			return;
		}

		self.data.step(dt);
		self.anim.step(dt);

		for (_, e) in &mut self.engine_anim {
			e.step(dt);
		}

		// Flare animations
		if !self.controls.thrust && self.last_controls.thrust {
			let flare = self.get_flare();
			if let Some(flare) = flare {
				if flare.engine_flare_on_stop.is_some() {
					for (_, e) in &mut self.engine_anim {
						e.jump_to(flare.engine_flare_on_stop.as_ref().unwrap());
					}
				}
			};
		} else if self.controls.thrust && !self.last_controls.thrust {
			let flare = self.get_flare();
			if let Some(flare) = flare {
				if flare.engine_flare_on_start.is_some() {
					for (_, e) in &mut self.engine_anim {
						e.jump_to(flare.engine_flare_on_start.as_ref().unwrap());
					}
				}
			};
		}
		// Update this now and not later,
		// since autopilot might change controls.
		self.last_controls = self.controls.clone();

		match self.data.get_state() {
			ShipState::Collapsing { .. } => {
				// Borrow checker hack, so we may pass self.data
				// to the collapse sequence
				let mut seq = self.collapse_sequence.take().unwrap();
				seq.step(dt, wrapper, new, &self.data, self.rigid_body, self.collider);
				self.collapse_sequence = Some(seq);

				if self.collapse_sequence.as_ref().unwrap().is_done() {
					self.data.finish_collapse();
				}
			}
			ShipState::Flying => {
				let controls = match self.autopilot {
					Some(x) => x.update_controls(dt, sf, PhysSimShipHandle(self.collider)),
					None => self
						.agent
						.update_controls(dt, sf, PhysSimShipHandle(self.collider)),
				};

				if let Some(controls) = controls {
					self.controls = controls;
				}

				if self.controls.landing {
					let controls = None;
					/*
					let controls = autopilot::auto_landing(
						res,
						img,
						PhysSimShipHandle(self.collider),
						Vector2::new(target.pos.x, target.pos.y),
					);*/

					// Try to land the ship.
					// True if success, false if failure.
					// Failure implies no state changes.
					let landed = 'landed: {
						let r = wrapper.get_rigid_body(self.rigid_body).unwrap();

						let t_pos = Vector2::new(target.pos.x, target.pos.y);
						let s_pos =
							Vector2::new(r.position().translation.x, r.position().translation.y);

						// TODO: deactivate collider when landing.
						// Can't just set_active(false), since we still need that collider's mass.

						// We're in land range...
						if (t_pos - s_pos).magnitude() > target.size / 2.0 {
							break 'landed false;
						}

						// And we'll stay in land range long enough.
						if (t_pos - (s_pos + r.velocity_at_point(r.center_of_mass()) * 2.0))
							.magnitude() > target.size / 2.0
						{
							break 'landed false;
						}

						let collider = wrapper.get_collider_mut(self.collider).unwrap();
						collider.set_collision_groups(InteractionGroups::new(
							Group::GROUP_1,
							Group::empty(),
						));
						self.data.start_land_on(target.clone());
						break 'landed true;
					};

					if landed {
						self.controls = ShipControls::new();
					}
				}

				self.step_physics(dt, wrapper);
				self.step_effects(dt, wrapper, new);

				// If we're firing, try to fire each gun
				if self.controls.guns {
					// TODO: don't allocate here. This is a hack to satisfy the borrow checker,
					// convert this to a refcell or do the replace dance.
					let pairs: Vec<(GunPoint, Option<Arc<Outfit>>)> = self
						.data
						.get_outfits()
						.iter_gun_points()
						.map(|(p, o)| (p.clone(), o.clone()))
						.collect();

					for (gun_point, outfit) in pairs {
						if self.data.fire_gun(&gun_point) {
							let outfit = outfit.unwrap();

							let mut rng = rand::thread_rng();

							let rigid_body = wrapper.get_rigid_body(self.rigid_body).unwrap();
							let ship_pos = rigid_body.translation();
							let ship_rot = rigid_body.rotation();
							let ship_ang = ship_rot.angle();
							let ship_vel =
								rigid_body.velocity_at_point(rigid_body.center_of_mass());

							let pos = ship_pos + (ship_rot * gun_point.pos);
							let gun = outfit.gun.as_ref().unwrap();

							let spread =
								rng.gen_range(-gun.projectile.angle_rng..=gun.projectile.angle_rng);
							let vel = ship_vel
								+ (Rotation2::new(ship_ang + spread)
									* Vector2::new(
										gun.projectile.speed
											+ rng.gen_range(
												-gun.projectile.speed_rng
													..=gun.projectile.speed_rng,
											),
										0.0,
									));

							let rigid_body = RigidBodyBuilder::kinematic_velocity_based()
								.translation(pos)
								.rotation(ship_ang)
								.linvel(vel)
								.build();

							let mut collider = match &gun.projectile.collider {
								ProjectileCollider::Ball(b) => ColliderBuilder::ball(b.radius)
									.sensor(true)
									.active_events(ActiveEvents::COLLISION_EVENTS)
									.build(),
							};

							collider.set_collision_groups(InteractionGroups::new(
								Group::GROUP_2,
								Group::GROUP_1,
							));

							let rigid_body = wrapper.insert_rigid_body(rigid_body);
							let collider = wrapper.insert_collider(collider, rigid_body);

							new.projectiles.push(ProjectileInstance::new(
								gun.projectile.clone(),
								rigid_body,
								self.data.get_faction().clone(),
								collider,
							));
						}
					}
				}
			}

			ShipState::UnLanding {
				to_position,
				current_z,
				from,
			} => {
				let controls = None;
				/*
				let controls = autopilot::auto_landing(
					&res,
					img,
					PhysSimShipHandle(self.collider),
					Vector2::new(to_position.translation.x, to_position.translation.y),
				);*/
				let r = wrapper.get_rigid_body_mut(self.rigid_body).unwrap();
				let max_d = (Vector2::new(from.pos.x, from.pos.y)
					- Vector2::new(to_position.translation.x, to_position.translation.y))
				.magnitude();
				let now_d = (r.translation()
					- Vector2::new(to_position.translation.x, to_position.translation.y))
				.magnitude();
				let f = now_d / max_d;

				let current_z = *current_z;
				let zdist = 1.0 - from.pos.z;

				if current_z <= 1.0 {
					// Finish unlanding ship
					self.data.finish_unland_to();
					wrapper
						.get_collider_mut(self.collider)
						.unwrap()
						.set_collision_groups(InteractionGroups::new(
							Group::GROUP_1,
							Group::GROUP_1 | Group::GROUP_2,
						));
				} else if current_z <= 1.5 {
					self.data
						.set_unlanding_z(1f32.max(current_z - (0.5 * dt) / 0.5));
					self.controls = ShipControls::new();
				} else {
					self.data.set_unlanding_z(1.0 - zdist * f);

					if let Some(controls) = controls {
						self.controls = controls;
					}
					self.step_physics(dt, wrapper);
				};
			}

			ShipState::Landing { target, current_z } => {
				let controls = None;
				/*
				let controls = autopilot::auto_landing(
					&res,
					img,
					PhysSimShipHandle(self.collider),
					Vector2::new(target.pos.x, target.pos.y),
				);
				*/

				let current_z = *current_z;
				let zdist = target.pos.z - 1.0;

				if current_z >= target.pos.z {
					// Finish landing ship
					self.data.finish_land_on();
					let r = wrapper.get_rigid_body_mut(self.rigid_body).unwrap();
					r.set_enabled(false);
					r.set_angvel(0.0, false);
					r.set_linvel(nalgebra::Vector2::new(0.0, 0.0), false);
				} else {
					self.data.set_landing_z(current_z + zdist * dt / 2.0);

					if let Some(controls) = controls {
						self.controls = controls;
					}
					self.step_physics(dt, wrapper);
				};
			}

			ShipState::Landed { .. } => {}

			ShipState::Dead => {
				wrapper.remove_rigid_body(self.rigid_body);
				self.is_destroyed = true;
			}
		};
	}

	/// Should this ship's data be removed?
	pub fn should_remove(&self) -> bool {
		self.is_destroyed
	}

	/// Update this frame's physics
	fn step_physics(&mut self, dt: f32, wrapper: &mut PhysWrapper) {
		let rigid_body = &mut wrapper.get_rigid_body_mut(self.rigid_body).unwrap();
		let ship_rot = rigid_body.rotation();
		let engine_force = ship_rot * (Vector2::new(1.0, 0.0) * dt);

		if self.controls.thrust {
			rigid_body.apply_impulse(
				vector![engine_force.x, engine_force.y]
					* self.data.get_outfits().get_stats().engine_thrust,
				true,
			);
		}

		if self.controls.right {
			rigid_body.apply_torque_impulse(
				self.data.get_outfits().get_stats().steer_power * -100.0 * dt,
				true,
			);
		}

		if self.controls.left {
			rigid_body.apply_torque_impulse(
				self.data.get_outfits().get_stats().steer_power * 100.0 * dt,
				true,
			);
		}
	}

	/// Spawn this frame's effects
	fn step_effects(&mut self, dt: f32, wrapper: &mut PhysWrapper, new: &mut NewObjects) {
		let rigid_body = wrapper.get_rigid_body(self.rigid_body).unwrap().clone();
		let collider = wrapper.get_collider(self.collider).unwrap().clone();

		let ship_content = self.data.get_content();
		let ship_pos = rigid_body.translation();
		let ship_rot = rigid_body.rotation();
		let ship_ang = ship_rot.angle();
		let mut rng = rand::thread_rng();

		if self.data.get_hull() <= ship_content.damage.hull {
			for e in &ship_content.damage.effects {
				if rng.gen_range(0.0..=1.0) <= dt / e.frequency {
					let pos = if let Some(pos) = e.pos {
						Vector2::new(pos.x, pos.y)
					} else {
						// Pick a random point inside this ship's collider
						let mut y = 0.0;
						let mut x = 0.0;
						let mut a = false;
						while !a {
							x = rng.gen_range(-1.0..=1.0) * ship_content.size / 2.0;
							y = rng.gen_range(-1.0..=1.0)
								* ship_content.size * ship_content.sprite.aspect
								/ 2.0;
							a = collider.shape().contains_local_point(&Point2::new(x, y));
						}
						Vector2::new(x, y)
					};

					let pos = ship_pos + (Rotation2::new(ship_ang) * pos);

					new.effects.push(EffectInstance::new(
						wrapper,
						e.effect.clone(),
						pos,
						self.rigid_body,
						None,
					));
				}
			}
		}
	}
}

/// Public mutable
impl ShipInstance {
	fn get_flare(&mut self) -> Option<Arc<Outfit>> {
		// TODO: better way to pick flare sprite
		for (h, _) in self.data.get_outfits().iter_outfits() {
			if h.engine_flare_sprite.is_some() {
				return Some(h.clone());
			}
		}
		return None;
	}

	/// Re-create this ship's engine flare animations
	/// Should be called whenever we change outfits
	fn update_flares(&mut self) {
		let flare = self.get_flare();
		if flare.is_none() {
			self.engine_anim = Vec::new();
			return;
		}

		self.engine_anim = self
			.data
			.get_content()
			.engines
			.iter()
			.map(|e| {
				(
					e.clone(),
					SpriteAutomaton::new(
						flare
							.as_ref()
							.unwrap()
							.engine_flare_sprite
							.as_ref()
							.unwrap()
							.clone(),
					),
				)
			})
			.collect();
	}

	/// Add one outfit to this ship
	pub fn add_outfit(&mut self, o: &Arc<Outfit>) {
		self.data.add_outfit(o);
		self.update_flares();
	}

	/// Add many outfits to this ship
	pub fn add_outfits(&mut self, outfits: impl IntoIterator<Item = Arc<Outfit>>) {
		for o in outfits {
			self.data.add_outfit(&o);
		}
		self.update_flares();
	}

	/// Remove one outfit from this ship
	pub fn remove_outfit(&mut self, o: &Arc<Outfit>) -> OutfitRemoveResult {
		let r = self.data.remove_outfit(o);
		self.update_flares();
		return r;
	}
}

/// Public immutable
impl ShipInstance {
	/// Get this ship's control state
	pub fn get_controls(&self) -> &ShipControls {
		&self.controls
	}

	/// Get this ship's engine animations
	pub fn iter_engine_anim(&self) -> impl Iterator<Item = &(EnginePoint, SpriteAutomaton)> {
		self.engine_anim.iter()
	}

	/// Get this ship's animation state
	pub fn get_anim_state(&self) -> AnimationState {
		self.anim.get_texture_idx()
	}

	/// Get this ship's game data struct
	pub fn get_data(&self) -> &ShipData {
		&self.data
	}
}
