use std::{
	collections::HashMap,
	sync::{
		atomic::{AtomicU64, Ordering},
		Arc,
	},
};

use galactica_content::{Content, Relationship};
use instance::{
	ship::{data::ShipState, ShipInstance},
	EffectInstance, ProjectileInstance,
};
use nalgebra::{Point2, Vector2};
use physwrapper::PhysWrapper;
use rapier2d::prelude::ColliderHandle;
use stateframe::{EffectStateframe, PhysSimShipHandle, ProjectileStateframe, SystemStateframe};

mod instance;
mod physwrapper;
mod shipagent;
mod stateframe;

pub(crate) struct NewObjects {
	pub projectiles: Vec<ProjectileInstance>,
	pub effects: Vec<EffectInstance>,
}

/// A list of new objects to create this frame
impl NewObjects {
	/// Create an empty NewObjects
	/// This should only be called once in each PhysSim
	pub fn new() -> Self {
		Self {
			projectiles: Vec::new(),
			effects: Vec::new(),
		}
	}

	/// Clear all lists in this struct.
	/// Call after saving all new objects.
	pub fn clear(&mut self) {
		self.projectiles.clear();
		self.effects.clear();
	}
}

/// A unique id given to each physics object
static PHYS_UID: AtomicU64 = AtomicU64::new(0);
fn get_phys_id() -> u64 {
	PHYS_UID.fetch_add(1, Ordering::Relaxed)
}

/// Simulates everything in one star system.
pub struct SystemSimulation {
	/// Game content
	ct: Arc<Content>,

	/// Rapier wrapper
	phys: PhysWrapper,
	/// Objects to create in the next frame
	new: NewObjects,

	effects: Vec<EffectInstance>,
	projectiles: HashMap<ColliderHandle, ProjectileInstance>,
	ships: HashMap<ColliderHandle, ShipInstance>,
}

impl SystemSimulation {
	pub fn collide_projectile_ship(
		&mut self,
		projectile_h: ColliderHandle,
		ship_h: ColliderHandle,
	) {
		let projectile = self.projectiles.get_mut(&projectile_h);
		let ship = self.ships.get_mut(&ship_h);
		if projectile.is_none() || ship.is_none() {
			return;
		}
		let projectile = projectile.unwrap();
		let ship = ship.unwrap();

		let r = projectile
			.faction
			.relationships
			.get(&ship.data.get_faction().index)
			.unwrap();
		let destory_projectile = match r {
			Relationship::Hostile => match ship.data.get_state() {
				ShipState::Flying { .. } => {
					ship.data.apply_damage(projectile.content.damage);
					true
				}
				ShipState::Collapsing { .. } => true,
				_ => false,
			},
			_ => false,
		};

		if destory_projectile {
			let pr = self.phys.get_rigid_body(projectile.rigid_body).unwrap();
			let v =
				pr.velocity_at_point(pr.center_of_mass()).normalize() * projectile.content.force;
			let pos = *pr.translation();
			let _ = pr;

			let r = self.phys.get_rigid_body_mut(ship.rigid_body).unwrap();
			r.apply_impulse_at_point(Vector2::new(v.x, v.y), Point2::new(pos.x, pos.y), true);

			// Borrow again, we can only have one at a time
			let pr = self.phys.get_rigid_body(projectile.rigid_body).unwrap();
			let pos = *pr.translation();

			match &projectile.content.impact_effect {
				None => {}
				Some(x) => {
					self.effects.push(EffectInstance::new(
						&mut self.phys,
						x.clone(),
						pos,
						projectile.rigid_body,
						Some(ship.rigid_body),
					));
				}
			};

			projectile.destroy_silent(&mut self.new, &mut self.phys);
		}
	}
}

impl SystemSimulation {
	pub fn new(content: Arc<Content>) -> Self {
		Self {
			ct: content,

			phys: PhysWrapper::new(),
			new: NewObjects::new(),

			effects: Vec::new(),
			projectiles: HashMap::new(),
			ships: HashMap::new(),
		}
	}

	/// Step this simulation by `dt` seconds.
	pub fn step(&mut self, dt: f32, sf: &SystemStateframe) {
		self.phys.step(dt);

		// Handle collision events
		while let Ok(event) = &self.phys.collision_queue.try_recv() {
			if event.started() {
				let a = event.collider1();
				let b = event.collider2();

				// If projectiles are part of this collision, make sure
				// `a` is one of them.
				let (a, b) = if self.projectiles.contains_key(&b) {
					(b, a)
				} else {
					(a, b)
				};

				let p = self.projectiles.get(&a);
				let s = self.ships.get_mut(&b);
				if p.is_none() || s.is_none() {
					continue;
				}
				self.collide_projectile_ship(a, b);
			}
		}

		// Step and garbage-collect projectiles
		self.projectiles.retain(|_, proj| {
			proj.step(dt, &mut self.new, &mut self.phys);
			!proj.should_remove()
		});

		// Step and garbage-collect ships
		self.ships.retain(|_, ship| {
			ship.step(dt, sf, &mut self.phys, &mut self.new);
			!ship.should_remove()
		});

		// Step and garbage-collect effects
		self.effects.retain_mut(|x| {
			x.step(dt, &mut self.phys);
			!x.is_destroyed()
		});

		// Process new objects
		for p in self.new.projectiles.iter() {
			self.projectiles.insert(p.collider, p.clone());
		}
		for e in self.new.effects.iter() {
			self.effects.push(e.clone());
		}
		self.new.clear();
	}

	/// Update an image with the current simulation state
	///
	/// We mutate (rather than create) a new [`SystemStateframe`]
	/// so we don't need to allocate memory every time we draw a frame.
	pub fn get_stateframe(&mut self, sf: &mut SystemStateframe) {
		sf.clear();

		for s in self.ships.values() {
			sf.ship_map
				.insert(PhysSimShipHandle(s.collider), sf.ships.len());
			sf.ships.push(s.get_stateframe(&self.phys))
		}

		for p in self.projectiles.values() {
			sf.projectiles.push(ProjectileStateframe {
				projectile: p.clone(),
				rigidbody: self.phys.get_rigid_body(p.rigid_body).unwrap().clone(),
			})
		}

		for e in self.effects.iter() {
			sf.effects.push(EffectStateframe {
				effect: e.clone(),
				rigidbody: self.phys.get_rigid_body(e.rigid_body).unwrap().clone(),
			})
		}
	}
}
