use galactica_content::Relationship;
use galactica_util::clockwise_angle;
use nalgebra::Vector2;

use crate::{
	instance::ship::data::ShipState,
	stateframe::{PhysSimShipHandle, SystemStateframe},
};

use super::{ShipAgent, ShipControls};

/// "Point" ship agent.
/// Point and shoot towards the nearest enemy.
#[derive(Debug, Clone)]
pub struct PointShipAgent {}

impl PointShipAgent {
	/// Create a new ship controller
	pub fn new() -> Self {
		Self {}
	}
}

impl ShipAgent for PointShipAgent {
	fn update_controls(
		&mut self,
		_dt: f32,
		sf: &SystemStateframe,
		this_ship: PhysSimShipHandle,
	) -> Option<ShipControls> {
		let mut controls = ShipControls::new();

		let my_ship = match sf.get_ship(&this_ship) {
			None => return None,
			Some(s) => s,
		};

		let this_rigidbody = &my_ship.rigidbody;
		let my_position = this_rigidbody.translation();
		let my_rotation = this_rigidbody.rotation();
		let my_angvel = this_rigidbody.angvel();
		let my_faction = my_ship.data.get_faction();

		// Iterate all possible targets
		let mut hostile_ships = sf.iter_ships().filter(
			// Target only flying ships we're hostile towards
			|s| match my_faction
				.relationships
				.get(&s.data.get_faction().index)
				.unwrap()
			{
				Relationship::Hostile => match s.data.get_state() {
					ShipState::Flying { .. } => true,
					_ => false,
				},
				_ => false,
			},
		);

		// Find the closest target
		let mut closest_enemy_position = match hostile_ships.next() {
			Some(s) => s.rigidbody.translation(),
			None => return Some(controls), // Do nothing if no targets are available
		};

		let mut d = (my_position - closest_enemy_position).magnitude();
		for s in hostile_ships {
			let p = s.rigidbody.translation();
			let new_d = (my_position - p).magnitude();
			if new_d < d {
				d = new_d;
				closest_enemy_position = p;
			}
		}

		let angle = clockwise_angle(
			&(my_rotation * Vector2::new(1.0, 0.0)),
			&(closest_enemy_position - my_position),
		);

		if angle < 0.0 && my_angvel > -0.3 {
			controls.right = true;
		} else if angle > 0.0 && my_angvel < 0.3 {
			controls.left = true;
		}

		controls.guns = true;
		return Some(controls);
	}
}
