//! Provides a snapshot of one frame of a physics simulation
//! that can then be passed to a renderer to be drawn.

use std::collections::HashMap;

use galactica_content::{EnginePoint, SpriteAutomaton};
use rapier2d::{dynamics::RigidBody, prelude::ColliderHandle};

use crate::instance::{
	ship::{data::ShipData, ShipCollapseSequence},
	EffectInstance, ProjectileInstance,
};

// TODO: replace with a more generic handle
/// A handle for a ship in this simulation
/// This lets other crates reference ships
/// without including `rapier2d`.
#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
pub struct PhysSimShipHandle(pub ColliderHandle);

/// A snapshot of one frame of a system simulation
#[derive(Debug)]
pub struct SystemStateframe {
	/// The ships in this frame
	pub(crate) ships: Vec<ShipStateframe>,

	/// Map ship handles to indices in ships
	pub(crate) ship_map: HashMap<PhysSimShipHandle, usize>,

	/// The projectiles in this frame
	pub(crate) projectiles: Vec<ProjectileStateframe>,

	/// The effects in this frame
	pub(crate) effects: Vec<EffectStateframe>,
}

impl SystemStateframe {
	/// Create an empty simulation image
	pub fn new() -> Self {
		Self {
			ships: Vec::new(),
			projectiles: Vec::new(),
			ship_map: HashMap::new(),
			effects: Vec::new(),
		}
	}

	/// Clear all buffers in this image
	pub(crate) fn clear(&mut self) {
		self.ship_map.clear();
		self.projectiles.clear();
		self.effects.clear();
		self.ships.clear();
	}

	/// Iterate ships in this image
	pub fn iter_ships(&self) -> impl Iterator<Item = &ShipStateframe> {
		self.ships.iter()
	}

	/// Get a ship by its handle
	pub fn get_ship(&self, handle: &PhysSimShipHandle) -> Option<&ShipStateframe> {
		self.ship_map.get(handle).map(|x| &self.ships[*x])
	}

	/// Iterate projectiles in this image
	pub fn iter_projectiles(&self) -> impl Iterator<Item = &ProjectileStateframe> {
		self.projectiles.iter()
	}

	/// Iterate effects in this image
	pub fn iter_effects(&self) -> impl Iterator<Item = &EffectStateframe> {
		self.effects.iter()
	}
}

/// A snapshot of a ship's state
#[derive(Debug)]
pub struct ShipStateframe {
	/// This ship's unique id
	pub uid: u64,

	/// This ship's game data
	pub data: ShipData,

	/// This ship's sprite animation state
	pub anim: SpriteAutomaton,

	/// Animation state for each of this ship's engines
	pub engine_anim: Vec<(EnginePoint, SpriteAutomaton)>,

	/// This ship's collapse sequence
	pub collapse_sequence: Option<ShipCollapseSequence>,

	/// The ship's rigidbody
	pub rigidbody: RigidBody,
}

/// a snapshot of a projectile's state
#[derive(Debug)]
pub struct ProjectileStateframe {
	/// The projectile's data
	pub projectile: ProjectileInstance,

	/// The projectile's rigidbody
	pub rigidbody: RigidBody,
}

/// a snapshot of a projectile's state
#[derive(Debug)]
pub struct EffectStateframe {
	/// The effect's data
	pub effect: EffectInstance,

	/// The effect's rigidbody
	pub rigidbody: RigidBody,
}

impl EffectStateframe {
	/// Get this effect's fade value
	pub fn get_fade(&self) -> f32 {
		let f = self.effect.fade;
		let l = self.effect.remaining_lifetime();
		return 1f32.min(l / f);
	}
}
